/*
#
#    This distribution makes makes asynchronous client/server communication possible 
#    using TLVs (Type Length Value). It achieves that letting the user define
#    an IDL which serves as input to a code generator.  The code generated is linked in
#    the client and the server program.
#
#    Copyright (C) <2005>  <Sanjay Jaiman>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/unistd.h>

#include <string.h>
#include <memory.h>
#include <Timer.h>

static char printbuff[256];

TimerElement::TimerElement(char *Name, unsigned int TOVal,  void (*callBack)(char *), bool recurring)
{
  timeVal = 0;  // Dynamic - used for figuring out next timeout
  timeoutVal = TOVal;  //Static - for this timerElement
  func = callBack;
  int len = strlen(Name);
  m_name = (char *)malloc(len + 1);
  memcpy(m_name, Name, len);
  m_name[len] = '\0';
  m_rucurring = recurring;
};

unsigned int Timer::getTimeInMilliSecs(struct timeval tp)
{
  return((tp.tv_sec * 1000) + ((tp.tv_usec/1000) % 1000));
}

TimerElement *Timer::addTimer(char *name, unsigned int timeVal,  void (*callBack)(char *), bool recurring)
{
  if (!lastTime) {
    lastTime = (struct timeval *)malloc(sizeof(struct timeval));
    gettimeofday(lastTime, 0);
  }
  TimerElement *tv = new TimerElement(name, timeVal, callBack, recurring);
  insertTimer(tv);
  return tv;
}

int Timer::deleteTimer(char *name)
{
  TimerElement *p = timerHead;
  for (; p ; p = p->next) {
    if (!strncmp(p->m_name, name, strlen(name))) {
      p = removeTimer(p);
      if (p) {
	delete p;
	free(p);
	return 1;
      }
      break;
    }
  }
  return 0;
}

int Timer::getNewTimeToWait(struct timeval &tp)
{
  unsigned int newTimeToWait = 0;
  if (!timerHead)
    return 0;
  newTimeToWait = timerHead->timeVal;
  //  sprintf(printbuff, "New Time to wait =  %d from timer %s\n", newTimeToWait, timerHead->m_name);
  //  printf(printbuff);
  tp.tv_sec = newTimeToWait / 1000;
  tp.tv_usec = (newTimeToWait % 1000)*1000;
  return 1;
}

void Timer::printTimerList(char *id)
{
  char *pbuff;
  TimerElement *t = NULL;

  pbuff = printbuff;
  pbuff += sprintf(pbuff, "Timer List %s:: ", id);
  for (t = timerHead; t; t = t->next) {
    pbuff += sprintf(pbuff, "%s/%d :: ", t->m_name, t->timeVal);
  }
  pbuff += sprintf(pbuff, "\n");
  printf(printbuff);
}

void Timer::timerFired()
{
  TimerElement *t = NULL;
  struct timeval currentTime;
  unsigned int ctime;
  unsigned int ltime;
  unsigned int elapsedTime;

  gettimeofday(&currentTime, 0);
  ctime = getTimeInMilliSecs(currentTime);
  ltime = getTimeInMilliSecs(*lastTime);
  elapsedTime = ctime - ltime;

  //  sprintf(printbuff, "%s Elapsed Time =  %d\n", Name, elapsedTime);
  //  printf(printbuff);
  *lastTime = currentTime;
  if (!timerHead)
    return;
  if (elapsedTime < timerHead->timeVal) {
    timerHead->timeVal -= elapsedTime;  // Decrement by elapsed time
    return;
  }

  unsigned int sigma = 0;
  Timer tempTimerlist;
  for (t = timerHead; t;) {
    TimerElement *next = t->next;
    sigma += t->timeVal;
    if (elapsedTime >= sigma) {
      t = removeTimer(t);  // Remove from main list
      t->func(t->m_name);  /* Callback */
      tempTimerlist.insertTimer(t);  // Insert in temp list
    }
    else {
      // Just fix the first one with (sigma > elapsed time) & break
      t->timeVal = sigma - elapsedTime;
      break;
    }
    t = next;
  }
  
  //  printTimerList("Orignal");
  //  tempTimerlist.printTimerList("Removed");

  for (t = tempTimerlist.getFirst(); t;) {
    TimerElement *next = t->next;
    if (t->m_rucurring)
      insertTimer(t);  // re-insert
    t = next;
  }
  //  printTimerList("Final");
}

void Timer::insertTimer(TimerElement *tv)
{
  TimerElement *p = NULL;
  unsigned int sum = 0;
  unsigned int thisTimeoutVal = tv->timeoutVal;
  tv->next = tv->prev = NULL;
  tv->timeVal = thisTimeoutVal;
  char *pbuff;
  
  if (!timerHead) {  // First one
    //sprintf(printbuff, "Inserting first timer %s\n", tv->m_name);
    //printf(printbuff);
    timerHead = timerTail = tv;
    return;
  }

  for (p = timerHead; p ; p = p->next) {
    if (sum + p->timeVal > thisTimeoutVal) {
      if (p->prev)
	p = p->prev;  // To insert after p
      break;
    }
    sum += p->timeVal;
  }
  
  pbuff = printbuff;
  if (!p) { // insert at end
    pbuff += sprintf(pbuff, "Inserting timer %s at the END\n", tv->m_name);
    tv->prev = timerTail;
    timerTail->next = tv;
    timerTail = tv;
  }
  else if (sum == 0) { // insert at top
    pbuff += sprintf(pbuff, "Inserting timer %s at the TOP\n", tv->m_name);
    tv->next = timerHead;
    timerHead->prev = tv;
    timerHead = tv;
  }
  else { // Insert after p
    pbuff += sprintf(pbuff, "Inserting timer %s after the timer %s\n", tv->m_name, p->m_name);
    tv->next = p->next;
    tv->prev = p;
    if (p->next)
      p->next->prev = tv;
    else
      timerTail = tv;
    p->next = tv;
  }
  tv->timeVal = thisTimeoutVal - sum;
  if (tv->next)
    tv->next->timeVal -= tv->timeVal;
  numtimers++;
  //printf(printbuff);
}

TimerElement *Timer::removeTimer(TimerElement *tv) /* returns the removed one */
{
  if (tv->prev) {tv->prev->next = tv->next;}
  if (tv->next) {tv->next->prev = tv->prev;}
  if (timerHead == tv) {timerHead = tv->next;}
  if (timerTail == tv) {timerTail = tv->prev;}
  numtimers--;
  tv->next = tv->prev = NULL;
  tv->timeVal = 0;
  return tv;
}
