/*
#
#    This distribution makes makes asynchronous client/server communication possible 
#    using TLVs (Type Length Value). It achieves that letting the user define
#    an IDL which serves as input to a code generator.  The code generated is linked in
#    the client and the server program.
#
#    Copyright (C) <2005>  <Sanjay Jaiman>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
*/

#include <stdio.h>
#include <iostream.h>
#include <stdlib.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <linux/sockios.h>
#include <netinet/in.h>
#include <net/if.h>
#include <signal.h>
#include <string.h>
#include <varargs.h>
#include <ctype.h>
#include <netdb.h>

#include "globals.h"
#include "Timer.h"

#include "myConn.h"

extern int h_errno;

unsigned long  subSystem_trace_mask = 0;
int trace_level= 0; /* default: zero, i.e., no trace output            */

static Connections clientConnections;

static void sig(int signo)
{
  if ((signo == SIGTERM) || (signo == SIGINT) ||
      (signo == SIGSEGV)|| (signo == SIGPIPE)) {

    printf ("\n");
      
    switch (signo) {
    case SIGSEGV: {
      printf ("Segmentation fault, aborting.\n");
      exit(1);
    }
    case SIGTERM: 
      // Put all you "clean_up" code here
      break;
    case SIGINT: {
      // Ignore
      printf("Received SIGINT - ignoring\n");
      exit(0);
    }
    case SIGPIPE: {
      // Ignore
      printf("Received SIGPIPE - ignoring\n");
    }
    }
  }
}


void init_signals()
{
  signal (SIGTERM, sig);
  signal (SIGINT, sig);
  signal (SIGSEGV, sig); 
  signal (SIGPIPE, sig); 
}	

void printUsage() {
  cout << "Usage:: client -p port [-s <source IP>]" << endl;
}

#define SERV_HOST_ADDR "127.0.0.1"

struct sockaddr_in serv_addr;

int verifyIpAddress(char *str)
{
  int i1, i2, i3, i4;
  if (sscanf(str, "%d.%d.%d.%d", &i1, &i2, &i3, &i4) == 4) {
    if (i1 >= 0 && i1 <= 255 &&
	i2 >= 0 && i2 <= 255 &&
	i3 >= 0 && i3 <= 255 &&
	i4 >= 0 && i4 <= 255)
      return 1;
  }
  return 0;
}

int server_port = -1;
char serverIp[80];
  
void connect()
{
  char errorString[128];
  if (SocketInterface::connectToServer(server_port, myConn::staticConn, serverIp, errorString) == 0) {
    printf("Connect to - %s, succeeded, %s\n", "127.0.0.1", errorString);
  }
  else {
    printf("Connect to - %s, not succeeded, %s\n", "127.0.0.1", errorString);
  }
}

void sendTestTlv();

void timer_routine1(char *timerName)
{
  printf ("User timer timer_routine() called for timer - %s ---\n", timerName);
  if (myConn::staticConn->isConnected()) {
    //    printf("\tConnected - Sending TLVs\n");
    //    for (int i = 0; i < 5; i++) {
    //      sendTestTlv();
    //    }
    // sendTestTlv();
    printf("\tConnected\n");
  }
  else {
    printf("\tNot Connected\n");
    connect();
  }
}


void timer_routine(char *timerName)
{
  printf ("User timer timer_routine() called for timer - %s ---\n", timerName);
  //  connect();
}

static bool debug = false;

int main (int argc, char* argv[])
{
  char ch;
  char *cp;

  strcpy(serverIp, "127.0.0.1");
  while ((ch = getopt(argc, argv, "dT:p:s:")) != EOF) {
    switch(ch) {
    case 'd':
      trace_level = 5;
      subSystem_trace_mask |= SOCKET_LIB_TRACE;
      debug = true;
      break;
    case 'p':
      server_port = atoi(optarg);
      break;
    case 's':
      strcpy(serverIp, optarg);
      break;
     case 'T': /* Tracing level */
       for (cp=optarg; *cp; cp++) /* test for numeric option */
         if (!isdigit((int)*cp)){
	   printf("Error: trace-level must be an unsigned integer\n");
	   printUsage();
	   exit(1);
	 }
       trace_level = atoi(optarg);
    }
  }
  if (server_port == -1) {
    printUsage();
    exit(1);
  }
  cout << "Client::Port = " << server_port << endl;

  init_signals();
  chdir("/");
  
  /* Timer stuff */
  Timer myTimer;
  myTimer.addTimer("connectTimer", 15000, timer_routine);  // 15 secs
  myTimer.addTimer("senndTestTlv", 1000, timer_routine1);  // 1 sec

  // Create "My" connection
  myConn::staticConn = new myConn(&clientConnections, CLIENT);

  connect();

  fd_set  readfds;
  fd_set  writefds;

  while (1) {
    struct timeval t;
    if (!myTimer.getNewTimeToWait(t)) {
      t.tv_sec = 1;  // 1 sec timeout - default
      t.tv_usec = 0;
    }
    int maxfdpl = 0;

    FD_ZERO(&readfds);
    FD_ZERO(&writefds);

    int tempMax = clientConnections.setReadWriteFdSets(readfds, writefds);
    maxfdpl = (maxfdpl > tempMax) ? maxfdpl : tempMax;

    maxfdpl++;

    // fd not used - could be used in read/write fd sets
    if (select(maxfdpl, (fd_set *)&readfds, (fd_set *)&writefds, (fd_set *)0, &t) < 0) {
      cout << "Error in select call" << endl;
      continue;  // we don't exit as we would like the signal handler to handle all exits
      //exit(0);
    }

    myTimer.timerFired();

    clientConnections.findConnectionAndDispatch(readfds, writefds);
    // cout << "Client::In select loop" << endl;
  }
  return 0;
}


void printDebug(char *buff, int buffLen)
{
  printf(buff);
}
