/* -------------------------------------------------------------------------- */
/*                                                                            */
/* Filename: main.c                                                           */
/*                                                                            */
/* Author:   Sanjay Jaiman                                                    */
/*           Amperion Inc.                                                    */
/*                                                                            */
/* Notice:   Copyright 2002, Amperion, Inc.  All rights reserved.             */
/*                                                                            */
/* Description: This file starts the Amperion agent and cli telnet server.    */
/*              The agent is based on WindRiver's Envoy/Emissary tools.       */
/*                                                                            */
/* Modification History:                                                      */
/*                                                                            */
/*                                                                            */
/* -------------------------------------------------------------------------- */

#include <stdio.h>
#include <iostream.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <signal.h>
#include <syslog.h>
#include "Timer.h"
#include "myConn.h"
#include <errno.h>

extern int errno;

unsigned long  subSystem_trace_mask = 0;
int trace_level= 0; /* default: zero, i.e., no trace output            */

char usage[]="Usage: %s [-p port]\n";

void exitgracefully()
{
  if (myConn::staticConn && myConn::staticConn->getFd() != -1) {
    ::close(myConn::staticConn->getFd());
  }

  exit(1);
}


static void sig(int signo)
{
  if ((signo == SIGTERM) || (signo == SIGINT) ||
      (signo == SIGSEGV || signo == SIGUSR1 || signo == SIGPIPE)) {

    switch (signo) {
    case SIGUSR1:
      {
	printf ("Caught SIGUSR1:: \n");
	break;
      }
    case SIGSEGV:
      {
	printf ("Caught SIGSEGV:: Segmentation fault, aborting.\n");
	exitgracefully();
      }
    case SIGTERM: 
      // Put all you "clean_up" code here
      printf ("Caught SIGTERM:: exiting\n");
      exitgracefully();
      break;
    case SIGINT:
      {
	// Ignore
	printf("Received SIGINT:: exiting\n");
	exitgracefully();
	break;
      }
    case SIGPIPE:
      {
	// Ignore
	// printf("Received SIGPIPE - ignoring\n");
	break;
      }
    }
  }
}

/*-----------------------------------------------------------------------*/

void init_signals()
{
  signal (SIGTERM, sig);
  signal (SIGINT, sig);
  signal (SIGSEGV, sig); 
  signal (SIGPIPE, sig); 
  signal (SIGUSR1, sig); 
}	

/*-----------------------------------------------------------------------*/

int connectToServer()
{
  struct sockaddr_in serv_addr;

  bzero((void *)&serv_addr, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = inet_addr("127.0.0.1");
  serv_addr.sin_port = htons(MYFIRST_PORT);
  
  //  printf("Trying to connect to port %d\n", myFirst_PORT);
  int sockFd = myConn::staticConn->getFd();
  if (sockFd == -1) {
    sockFd = SocketInterface::createNonBlockingSocket();
    myConn::staticConn->setFd(sockFd);
  }
  return myConn::staticConn->Connect((struct sockaddr *)&serv_addr);
}

/*-----------------------------------------------------------------------*/

void timer_routine1(char *name)
{
  //  printf("*********** User timer routine %s fired *********** \n", name);
  if (!myConn::staticConn->isConnected()) {
    if (connectToServer()) {
      printf("Connect to myFirst succeeded\n");
    }
  }
}

/*-----------------------------------------------------------------------*/
static Connections myConnections;


int main(int argc, char *argv[])
{
  int     portnum;
  int     c;
  char   *cp;
  Timer  myTimer("AgentMain");
  
  opterr = 0;     /* Keep getopt quiet */
  while ((c = getopt(argc, argv, "p:")) != EOF) {
    switch((char)c)  {
    case 'p': /* port number */
      for (cp=optarg; *cp; cp++) /* test for numeric option */
	if (!isdigit((int)*cp)){
	  printf("Error: port number must be an unsigned integer\n");
	  printf(usage,argv[0]);
	  exit(1);
	}
      portnum = atoi(optarg);
      break;
    case '?': /* Unknown */
    default:
      printf("Error: unknown argument\n");
      printf(usage,argv[0]);
      exit(1);
    }
  }
  
  if (optind<argc)  {/* additional (invalid) args on the commandline */
    printf("Error: extra argument encounted which does not begin with a dash\n");
    printf(usage,argv[0]);
    exit(1);
  }
     
  init_signals();
  /* Timer stuff */
  myTimer.addTimer("ConnectTimer", 15000, timer_routine1); // every 15 seconds
  //myTimer.printTimerList("Main");

  // Create "My" connection
  myConn::staticConn = new myConn(&myConnections, CLIENT);
  printf("Agent:: FD of myFirst conn = %d\n", myConn::staticConn->getFd());
  if (connectToServer() && myConn::staticConn->isConnected()) {
    printf("Connect to myFirst succeeded\n");
  }

  while(1) {
    struct timeval tp;
    
    if (!myTimer.getNewTimeToWait(tp)) {
      tp.tv_sec = 1;  // 1 sec timeout - default
      tp.tv_usec = 0;
    }

    int maxfdpl = 0;
    fd_set  readfds;
    fd_set  writefds;

    FD_ZERO(&readfds);
    FD_ZERO(&writefds);

    int maxConnFd = 0;
    int myFirstFd = myConn::staticConn->getFd();

    if (myConn::staticConn->isConnectPending()) {
      FD_SET(myFirstFd, &readfds);
      FD_SET(myFirstFd, &writefds);
      maxfdpl = (maxfdpl < myFirstFd) ? myFirstFd : maxfdpl;
    }
    else {
      if (myFirstFd != -1) {
	FD_SET(myFirstFd, &readfds);
	maxfdpl = (maxfdpl < myFirstFd) ? myFirstFd : maxfdpl;
      }
    }
    
    maxfdpl++;
    //    printf("Main Thread - entering select call :: tp = %d:%d :: maxfdpl = %d\n", tp.tv_sec, tp.tv_usec, maxfdpl);
    //    myTimer.printTimerList("Main Timer List");

    if (select(maxfdpl, (fd_set *)&readfds, (fd_set *)&writefds, (fd_set *)0, &tp) < 0) {
      perror("Select failed: ");
      continue;  // we don't exit as we would like the signal handler to handle all exits
    }
    
    myConnections.findConnectionAndDispatch(readfds, writefds);
    myTimer.timerFired();
  } /* while (1) */
  exitgracefully();
}


void printDebug(char *buff, int buffLen)
{
}
