/*
#
#    This distribution makes makes asynchronous client/server communication possible 
#    using TLVs (Type Length Value). It achieves that letting the user define
#    an IDL which serves as input to a code generator.  The code generated is linked in
#    the client and the server program.
#
#    Copyright (C) <2005>  <Sanjay Jaiman>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
*/

/*
  _############################################################################
  _## 
  _##  SocketInterface.h
  _## 
  _##
  _##  
  _##########################################################################
*/

#ifndef _TLV_H_
#define _TLV_H_

#include <stdio.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <syslog.h>

#define TLSIZE (2*sizeof(int))


class ipaddress {
 public:
  ipaddress() {val = 0;};
  ipaddress(char *str) {val = getVal(str);};
  ipaddress(unsigned long v) {val = v;};

 private:
  unsigned int val;

 private:
  unsigned int getVal(const char *str) {
    unsigned int i1, i2, i3, i4;
    unsigned int Val = 0;
    int num = sscanf(str, "%u.%u.%u.%u", &i1, &i2, &i3, &i4);
    if (num == 4 &&
	i1 >= 0 && i1 <= 255 &&
	i2 >= 0 && i2 <= 255 &&
	i3 >= 0 && i3 <= 255 &&
	i4 >= 0 && i4 <= 255)
      Val = i1 << 24 | i2 << 16 | i3 << 8 | i4;
    return Val;
  };

 public:
  unsigned int getIntVal() {return val;};
  void setVal(unsigned int v) {val = v;};
  void print() {printf("%s", charVal());};
  // This function returns a pointer to a static buffer, thus strcpy needs to be done after haveing called it
  char *charVal() {
    struct in_addr ip_ad;
    ip_ad.s_addr = val;
    return inet_ntoa(ip_ad);
  };
  ipaddress &operator=(char *&ipStr) {
    val = getVal(ipStr);
    return *this;
  };
  ipaddress &operator=(in_addr &in) {
    val = in.s_addr;
    return *this;
  };
};

// Use malloc & free in this class because "realloc" used on the String
class StringSock {
 public:
  StringSock() {String = NULL; Length = 0;};
  StringSock(char *str, int len) {
    String = NULL; Length = 0; if (!len) return;
    Length = len+1;
    String = (char *)malloc(Length);
    if (!String) {
      cout << "tlv.h:: StringSock 1:: FATAL ERROR:: Error in malloc" << endl;
      syslog(LOG_CRIT, "socketInterface:tlv:Unable to allocate memory - exiting");
      exit(0);
    }
    memcpy(String,str,len);String[len] = '\0';};
  ~StringSock() {free(String);};
 public:
  unsigned int Length;
  char *String;
 public:
  char *putString(char *str) {
    if (!str) return NULL;
    int l = strlen(str); String = (char *)realloc(String,l+1);
    if (String == NULL) {
      cout << "tlv.h:: StringSock 1:: FATAL ERROR:: Error in malloc" << endl;
      syslog(LOG_CRIT, "socketInterface:tlv:Unable to allocate memory - exiting");
      exit(0);
    }
    memcpy(String, str, l); String[l] = '\0';
    Length = l+1; return String;}
  // This one should be used for transporting binary data
  char *putString(char *str, int len) {
    if (!str || len == 0) return NULL;
    String = (char *)realloc(String, len); 
    if (!String) {
      cout << "tlv.h:: StringSock 2:: FATAL ERROR:: Error in malloc" << endl;
      syslog(LOG_CRIT, "socketInterface:tlv:Unable to allocate memory - exiting");
      exit(0);
    }
    memcpy(String, str, len);Length = len; return String;}
  int  sprint(char *buff) {getString(buff); buff[Length] = '\0'; return Length+1;};
  void print() {char buff[Length+1]; bzero(buff, Length+1); getString(buff); printf(buff);};
  void getString(char *str) {snprintf(str, Length+1, "%s", String);}; // 1 for the additional '\0'
  StringSock &operator=(char *str) {
    putString(str);
    return *this;
  };
  StringSock &operator=(StringSock &in) {
    if (in.Length) {
      putString(in.String, in.Length);
    }
    return *this;
  };
  bool operator==(StringSock &in) {
    if (in.Length == Length) {
      if (!strncmp(in.String, String, Length)) {
	return true;
      }
    }
    return false;
  };
  bool operator==(char *in) {
    unsigned int len = strlen(in);
    if (len == Length) {
      if (!strncmp(in, String, Length)) {
	return true;
      }
    }
    return false;
  };
};

class tlv {
 public:
  // TLVs use the value from user space - user is to allocate and free the space
  tlv(){
    State = TLV_TYPE_STATE;
    Length = Type = 0;  //unknown
    Value = NULL;
  };
  tlv(int type, int length, char *val){
    Type = type;
    Length = length; 
    Value = (char *)malloc(Length);
    if (!Value) {
      cout << "tlv.h:: tlv: FATAL ERROR:: Error in malloc" << endl;
      syslog(LOG_CRIT, "socketInterface:tlv:Unable to allocate memory - exiting");
      exit(0);
    }
    memcpy(Value, val, Length);
    State = TLV_COMPLETE;
  };
  
  ~tlv() {flush();};

  void reinit() {
    State = TLV_TYPE_STATE;
  }
  void flush() {
    reinit(); Length = Type = 0; if (Value) free(Value); Value = NULL;
  };
  
  static int getLengthOfTLV(int valLength) {return valLength + 2*sizeof(int);};

  enum {
    TLV_COMPLETE = 0,
    TLV_TYPE_STATE,
    TLV_LEN_STATE,
    TLV_VAL_STATE
  } tlv_state;

 private:
  int State;
  int Type;
  unsigned int Length;
  char *Value;

 public:
  int getLength(){return Length;};
  int getType(){return Type;};
  char *getVal(){return Value;};
  void print() {
    cout << "Printing TLV::"<< endl;
    cout << "  Type = " << Type << endl;
    cout << "  Length = " << Length << endl;
    char p[Length+1];
    memcpy(p, Value, Length);
    p[Length]='\0';
    printf("  Value = %s\n", p);
  };
  int build(char *&buf, unsigned int &bufLength);
  int getState() {return State;};
  int extract(tlv *&out, char *&buf, unsigned int &bufLength);

 public:
  // Decode functions
  static float getFloat(char *&buf) {
    union {
      unsigned int I;
      float F;
    }U;
    U.I = getUnsignedInt(buf);
    return U.F;
  }
  static unsigned int getUnsignedInt(char *&buf) {
    unsigned int val = 0;
    for (int i = 0; i < 4; i++) {
      val = val << 8;
      val |= (0x000000ff & *buf++);
    }
    return val;
  };
  static int getInt(char *&buf) {
    int val = 0;
    for (int i = 0; i < 4; i++) {
      val = val << 8;
      val |= (0x000000ff & *buf++);
    }
    return val;
  };
  static int decodeTL(char *&buf, int &lengthOfBuffer, int &type, int &length);
  static int decodeINT(char *buf, int lengthOfBuffer, int T, int &V);
  static int decodeUINT(char *buf, int lengthOfBuffer, int T, unsigned int &V);
  static int decodeIpaddress(char *buf, int lengthOfBuffer, int T, ipaddress &V);
  static int decodeFLOAT(char *buf, int lengthOfBuffer, int T, float &V);
  static int decodeSTRING(char *buff, int lengthOfBuffer, int T, StringSock &V);

  // Encode functions
  static int encodeTL(char *&buf, int type, int length);
  static int encodeBuffer(char *&buf, int type, int length, char *val);

  static void putInt(char *&buff, int val) {
    for (int i = 0; i < 4; i++) {
      *buff++ = (val & 0xff000000) >> 24;
      val = val << 8;
    }
  }
  static void putUnsignedInt(char *&buff, unsigned int val) {
    for (int i = 0; i < 4; i++) {
      *buff++ = (val & 0xff000000) >> 24;
      val = val << 8;
    }
  }
  static void putFloat(char *&buff, float val) {
    union {
      unsigned int I;
      float F;
    }U;
    U.F = val;
    putUnsignedInt(buff, U.I);
  }
  static void putIntTLV(char *&buff, int T, int V) {
    encodeTL(buff, T, 4);
    putInt(buff, V);
  }
  static void putUIntTLV(char *&buff, int T, unsigned int V) {
    encodeTL(buff, T, 4);
    putUnsignedInt(buff, V);
  }
  static void putFloatTLV(char *&buff, int T, float V) {
    encodeTL(buff, T, 4);
    putFloat(buff, V);
  }
  static void putStringTLV(char *&buff, int T, StringSock &V) {
    encodeBuffer(buff, T, V.Length, V.String);
  }
  static void putIpaddressTLV(char *&buff, int T, ipaddress &V) {
    unsigned int v = V.getIntVal();
    v = htonl(v);
    putUIntTLV(buff, T, v);
  }
};


#endif
